/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE

#include <asm/unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/aio_abi.h>
#include <assert.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

static inline int io_setup(unsigned nr_events, aio_context_t *ctxp)
{
	return syscall(__NR_io_setup, nr_events, ctxp);
}

static inline int io_getevents(aio_context_t ctx, long min_nr, long nr,
			       struct io_event *events,
			       struct timespec *timeout)
{
	return syscall(__NR_io_getevents, ctx, min_nr, nr, events, timeout);
}

static inline int io_submit(aio_context_t ctx, long nr, struct iocb **iocbpp)
{
	return syscall(__NR_io_submit, ctx, nr, iocbpp);
}

#if __GLIBC_PREREQ(2, 8)
#include <sys/eventfd.h>
#elif __GLIBC_PREREQ(2, 7)
/* they are defined but the header file is missing */
#include <stdint.h>
typedef uint64_t eventfd_t;
extern int eventfd (int __count, int __flags) __THROW;
extern int eventfd_read (int __fd, eventfd_t *__value);
extern int eventfd_write (int __fd, eventfd_t value);
#else
#error Unsupported version.
#endif

int main(int argc, char *argv[])
{
	int err;
	aio_context_t laio;
	eventfd_t evc;
	struct io_event ev;
	struct iocb *pcb, laiocb = {
		.aio_lio_opcode	= IOCB_CMD_PREAD,
		.aio_nbytes	= BUFSIZ,
		.aio_offset	= 0,
		.aio_flags	= IOCB_FLAG_RESFD
	};

	laiocb.aio_fildes = open(argv[1], O_RDONLY | O_DIRECT);
	assert(laiocb.aio_fildes >= 0);
	laiocb.aio_data = laiocb.aio_fildes;
	err = posix_memalign((void *)&laiocb.aio_buf, 512, BUFSIZ);
	assert(!err);
	laiocb.aio_resfd = eventfd(0, 0);
	assert(laiocb.aio_resfd >= 0);

	memset(&laio, 0, sizeof(laio));
	err = io_setup(1, &laio);
	assert(!err);
	pcb = &laiocb;
	err = io_submit(laio, 1, &pcb);
	assert(err == 1);

	/* do other jobs */

	err = eventfd_read(laiocb.aio_resfd, &evc);
	assert(!err);
	assert(evc == 1);
	err = io_getevents(laio, 0, 1, &ev, NULL);
	assert(err == 1);
	assert(ev.obj == (unsigned long)&laiocb);

	printf("%s: io_submit from fd %llu completed\n",
	       __func__, ev.data);
	printf("%s, %lld read\n", argv[1], ev.res);

	return 0;
}
