/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#define _REENTRANT

#include <sys/stat.h>
#include <sys/types.h>
#include <aio.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "laiotest.h"

int do_test(char *fname, sigevent_t *sev, int expected)
{
	int err, nio;
	struct aiocb cb;
	char *p;
	struct stat st;

	nio = 0;
	err = posix_memalign((void *)&p, 512, BUFSIZ);
	assert(!err);
	cb.aio_fildes = open(fname, O_RDONLY | LaioFlag);
	assert(cb.aio_fildes >= 0);
	err = fstat(cb.aio_fildes, &st);
	assert(!err);
	assert(st.st_size <= BUFSIZ);

	cb.aio_reqprio = 0;
	cb.aio_buf = p;
	cb.aio_nbytes = BUFSIZ;
	cb.aio_offset = 0;
	cb.aio_sigevent = *sev;
	err = aio_read(&cb);
	assert(!err);
	nio++;

	do_cancel(&cb, st.st_size);

	wait_for_all_notification(sev, expected);

	err = laio_fin();
	assert(!err);

	return nio;
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	int nio, n;
	sigevent_t sev;

	init_test();

	set_none(&sev);
	nio = do_test(argv[1], &sev, 0);

	set_sig(&sev);
	n = do_test(argv[1], &sev, nio);
	assert(n == nio);
	assert(n == nnotify);

	set_thread(&sev);
	n = do_test(argv[1], &sev, nio);
	assert(n == nio);
	assert(n == nnotify);

	return 0;
}
