#!/bin/sh

# Copyright (C) 2010 J. R. Okajima

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

src=/usr/share
nread=10
mntpnt=/dev/shm/decomp

set -eu
types="ext2 sq40 sq40nofrag sq40para sq40nofrag_para cloop sq40nested \
	sq40lzma sqlzma34"
if [ $# -ne 1 ]
then
	cat <<- EOF 1>&2
	Usage: $0 'TYPE|clean'
	For TYPE, choose one among these.
	  $types
	For sq40para and sq40nofrag_para, see sq40para.mk.
	For cloop, see cloop.mk.
	For sq40lzma, use linux-2.6.34-rc6 and see sq30lzma.mk.
	For sqlzma34, use linux-2.6.27.4 and see sqlzma34.mk.
	EOF
	false
fi

base=base.img
alias time="time -f %e"
alias stat="stat -c '%B x %b, %s, %n'"
alias make="make Mntpnt=$mntpnt NRead=$nread"

if [ $1 = clean ]
then
	echo $types |
	tr ' ' '\n' |
	sed -e 's/$/.img/' |
	xargs rm -f
	rm -f $base
	make -s -f decomp_gen.mk clean
	exit 0
fi

mount2()
{
	cd
	sudo umount $mntpnt
	do_mount2
	cd $mntpnt
}


mksrc()
{
	echo $src |
	tr -s '\n' |
	sed -e s:^:${mntpnt}: |
	xargs -r mkdir -p
	for i in $src
	do sudo mount -o bind $i $mntpnt/$i
	done
}

rmsrc()
{
	for i in $src
	do sudo umount $mntpnt/$i
	done
}

mount_ext2()
{
	mksrc
	echo $(du -s $mntpnt) \
		$(stat -f -c '%S x %b, %T, %n' $mntpnt)
	alias mount2=true
}

umount_ext2()
{
	rmsrc
}

MksqOpt="-no-progress -noappend"
mksq_cmd() { $PWD/sq4.0.wcvs/squashfs/squashfs-tools/mksquashfs $@; }
mksq()
{
	local src dst
	src="$1"
	dst="$2"
	shift 2
	test -f ${dst} && return 0
	mksq_cmd ${src} ${dst} ${MksqOpt} $@
}

do_sq() # img
{
	stat $1
	eval "do_mount2() { sudo mount -t squashfs -o ro,loop ${1} ${mntpnt}; }"
	do_mount2
}

mount_sq40() # img opts
{
	local dst=sq40
	if [ $# -ge 1 ]
	then
		dst=$1
		shift
	fi
	dst=$cwd/$dst.img

	mksrc
	mksq $mntpnt $dst $@
	rmsrc
	do_sq $dst
}

umount_sq40()
{
	sudo umount $mntpnt
	sudo modprobe -r squashfs
	sudo modprobe -r zlib_inflate
}

mount_sq40nofrag()
{
	mount_sq40 sq40nofrag -no-fragments
}

umount_sq40nofrag()
{
	umount_sq40
}

mount_sq40para()
{
	sudo modprobe zlib_inflate
	sudo insmod ${cwd}/sq40para/squashfs.ko
	mount_sq40
}

umount_sq40para()
{
	umount_sq40
}

mount_sq40nofrag_para()
{
	sudo modprobe zlib_inflate
	sudo insmod ${cwd}/sq40para/squashfs.ko
	mount_sq40nofrag
}

umount_sq40nofrag_para()
{
	umount_sq40nofrag
}

mkbase()
{
	test -f $base && return 0

	local mega=$(echo $(($(du -sm ${src} | cut -f1) + 1)))

        dd if=/dev/zero of=$base bs=1M count=$mega
	mkfs -t ext2 -q -F -m 0 -T small $base
	sudo mount -o loop $base $mntpnt
	sudo chmod o+w $mntpnt
	{ tar -cf - ${src} 2> /dev/null | tar -C $mntpnt -xpf -; }
	#find $mntpnt | wc -l
	sudo chmod o-w $mntpnt
	sudo umount $mntpnt
}

mount_cloop()
{
	local img=cloop.img

	if [ ! -f $img ]
	then
		mkbase
		$cwd/cloop-2.636/create_compressed_fs -q $base $img
		sudo chown $USER $img
	fi
	stat $img
	sudo modprobe cloop
	sleep 1
	sudo losetup -r /dev/cloop0 $img
	eval "do_mount2() { sudo mount -o ro /dev/cloop0 $mntpnt; }"
	do_mount2
}

umount_cloop()
{
	sudo umount $mntpnt
	sudo modprobe -r cloop zlib_inflate
}

mount_sq40nested()
{
	local img=$cwd/sq40nested.img

	if [ ! -f $img ]
	then
		mkbase
		mksq $base $img
	fi
	do_sq $img
	eval "do_mount2() { sudo mount -o ro,loop ${mntpnt}/${base} ${mntpnt}; }"
	do_mount2
}

umount_sq40nested()
{
	sudo umount $mntpnt
	sudo umount $mntpnt
	sudo modprobe -r squashfs
	sudo modprobe -r zlib_inflate
}

mount_sq40lzma()
{
	mount_sq40 sq40lzma -comp lzma
}

umount_sq40lzma()
{
	umount_sq40
	sudo modprobe -r decompress_unlzma
}

mount_sqlzma34()
{
	eval "mksq_cmd() { ${cwd}/sqlzma3.4-457/squashfs3.4/squashfs-tools/mksquashfs \$@; }"
	mount_sq40 sqlzma34
}

umount_sqlzma34()
{
	sudo umount $mntpnt
	sudo rmmod squashfs sqlzma unlzma
	sudo modprobe -r zlib_inflate
}

mount_sqlzma34nolzma()
{
set -x
	eval "mksq_cmd() { ${cwd}/sqlzma3.4-457/squashfs3.4/squashfs-tools/mksquashfs \$@; }"
	mount_sq40 sqlzma34nolzma -nolzma
set +x
}

umount_sqlzma34nolzma()
{
	umount_sqlzma34
}

########################################

cd "$(dirname $0)"
cwd="$PWD"
cd "$OLDPWD"

mkdir -p $mntpnt
mount_$1

echo -n generating... ''
make -s -f ${cwd}/decomp_gen.mk
echo done

########################################
# here we go
uname -a
cd $mntpnt
#df .

tmp=/tmp/$$
rc() # type
{
	#echo 3 | sudo dd of=/proc/sys/vm/drop_caches 2> /dev/null
	mount2
	echo $1 x $nread
	vmstat -n 1 > $tmp &
	sleep 1
	time ${cwd}/decomp_$1.sh
	kill $!
	sed -e 's/^/raw: /' $tmp

	tail -n +4 $tmp |
	awk -v nf=16 -v nl=$(cat $tmp | wc -l) '
	{
		for (i = 1; i <= nf; i++)
			a[i] += $i;
	}
	END {
		if (FNR != nl - 3)
			print "something is wrong,", FNR, ",", nl;
		# for (i = 1; i <= nf; i++)
		# 	printf a[i] " ";
		# print "\n";
		for (i = 1; i <= nf; i++)
			printf("%.1f ", a[i] / FNR);
		printf "\n";
	}
	'
	rm $tmp
}

for i in seq rand
do
	for j in "" _para
	do rc $i$j
	done
done

rm -f $tmp $tmp.*

cd
umount_$1
